package cn.ischoolcode.javabook.dushichat;

import java.awt.*;
import java.awt.event.*;
import java.io.*;

/**
 * 运行在客户端的文件传输线程。这个类继承
 * 了DushiFileTransfer类。通过该服务器。作为发件人，它
 * 读取和上传文件块到服务器。作为接收器，它接受和
 * 保存从客户端发送传入的文件块，通过该服务器。 
 * 
 * @author zcj
 */
public class DushiClientFileTransfer extends DushiFileTransfer {
	// 这个类管理客户端用户的文件传输（接收部分）。

	private static final long serialVersionUID = 220L;
	private Class thisClass = DushiClientFileTransfer.class;
	private DushiPanel parent = null;
	private DushiUser[] toUsers = null;
	private boolean sender = false;
	private SenderDialog senderDialog = null;
	private boolean cancelled = false;
	protected DushiUser fromUser = null;
	protected boolean accepted = false;

	public DushiClientFileTransfer(DushiPanel p, DushiUser[] u,
		RandomAccessFile f, String fn, long ts, int tb, int bs, int minutes) {
		// 这是我们启动文件传输时的构造函数
		// （即我们是发件人）

		  super(fn, ts, tb, bs, minutes);

		  parent = p;
		  toUsers = u;
		  transFile = f;
		  sender = true;

		  start();
	   }

	   public DushiClientFileTransfer(DushiPanel p, DushiUser fu, String fn,
			long ts, int tb, int bs, int minutes) {
		   // 这是当我们被要求接收文件传输时的构造函数。
		

		   super(fn, ts, tb, bs, minutes);

		   parent = p;
		   fromUser = fu;
		   sender = false;

		   start();
	   }

	   public void run() {
		   if (sender)
			  runSender();
		   else
			  runReceiver();

		   parent.connection.fileTransfers.remove(fileName);
	   }

	private void runSender() {
		//当我们的用户请求将文件传输给一个或多个其他用户时，就会运行此线程。
		//它将初始文件传输请求发送到服务器，然后读取并发送文件的块。之后，
		//它简单地显示进度以及其他用户是否接受了文件传输。

		byte[] data = null;

		senderDialog = new SenderDialog(parent, fileName, toUsers);
		senderDialog.setVisible(true);

		// 告诉服务器我们要发送一个文件
		parent.connection.sendFile(fileName, totalSize, totalBlocks, blockSize,
				expireMins);

		senderDialog.progress(0);

		// 现在立即将文件的块发送到服务器
		for (int count = 0; count < totalBlocks; count++) {
			try {
				data = getFileBlock(count);
			} catch (IOException e) {
				senderDialog.dispose();
				new DushiOkDialog(parent.parentWindow,
						parent.strings.get("error"), parent.strings.get(
								thisClass, "cantread"),
						parent.strings.get("ok"), false).setVisible(true);
				cancel();
			}

			parent.connection.sendFileBlock(fileName, count, data);
			senderDialog.progress((count * 100) / totalBlocks);
		}
		senderDialog.progress(100);

		try {
			transFile.close();
		} catch (IOException f) {
		}

		// 没有什么可做的了，但我们希望在用户关闭之前保持线程运行
		while (!cancelled && senderDialog.isVisible()) {
			try {
				sleep(100);
			} catch (InterruptedException e) {
			}
		}

		senderDialog.dispose();
	}

	private void runReceiver() {
		//当另一个用户（即我们的用户是预期的接收者）请求文件传输时，
		//该线程会运行。它询问用户是否接受它，在哪里保存它，等等。
		//然后它跟踪进度，并在最后让用户知道传输是否成功。

		String[] choiceStrings = new String[3];
		DushiChoiceDialog choiceDialog = null;
		choiceStrings[0] = parent.strings.get("yes");
		choiceStrings[1] = parent.strings.get("no");
		choiceStrings[2] = parent.strings.get("ignore");
		FileDialog fileDialog = null;
		File destFile = null;
		DushiProgressDialog progressDialog = null;

		// 首先，确定用户是否要接受文件传输
		choiceDialog = new DushiChoiceDialog(
				parent.parentWindow,
				parent.strings.get(thisClass, "filetrans"),
				(parent.strings.get("user") + " " + fromUser.name + " "
						+ parent.strings.get(thisClass, "request") + " "
						+ fileName + " (" + totalSize + "b).\n" + parent.strings
						.get(thisClass, "accept")), choiceStrings, 0, false);
		choiceDialog.setVisible(true);

		// 现在我们等待，直到用户选择了一些东西。
		while (choiceDialog.isVisible() && !cancelled) {
			try {
				sleep(100);
			} catch (InterruptedException e) {
			}
		}

		choiceDialog.dispose();

		if (cancelled)
			return;

		switch (choiceDialog.choice) {
		case 0:
			accepted = true;
			break;
		case 1:
		default:
			accepted = false;
			break;
		case 2:
			parent.connection.addToIgnored(fromUser.name);
			accepted = false;
			return;
		}

		// 如果传输被接受，请立即使用一个临时文件启动它，我们稍后将对其进行重命名。
		if (accepted) {
			try {
				tmpFile = File.createTempFile(fileName, null);
				transFile = new RandomAccessFile(tmpFile, "rw");
			} catch (IOException e) {
				new DushiOkDialog(
						parent.parentWindow,
						parent.strings.get("error"),
						(parent.strings.get(thisClass, "cantwrite") + " " + tmpFile
								.getName()), parent.strings.get("ok"), false)
						.setVisible(true);
				return;
			}
		}

		// 发送关于我们是否接受该文件的通知
		parent.connection.sendFileAccept(fromUser.id, fileName, accepted);

		if (!accepted)
			return;

		// 现在我们需要找到他们想要保存的位置。启动一个文件对话框，
		// 让用户选择文件位置
		fileDialog = new FileDialog(parent.parentWindow, parent.strings.get(
				thisClass, "savefile"), FileDialog.SAVE);
		parent.centerDialog(fileDialog);
		fileDialog.setFile(fileName);
		fileDialog.setVisible(true);

		if (fileDialog.getFile() != null)
			destFile = new File(fileDialog.getDirectory()
					+ fileDialog.getFile());
		else
			// 未选择任何内容，取消。
			cancel();

		progressDialog = new DushiProgressDialog(parent.parentWindow,
				parent.strings.get(thisClass, "filetrans"),
				(parent.strings.get(thisClass, "receiving") + " " + fileName),
				parent.strings.get("dismiss"), false);
		progressDialog.setVisible(!cancelled);

		while (!receivedAll && !cancelled) {
			progressDialog.progress((receivedBlocks * 100) / totalBlocks);
			try {
				sleep(100);
			} catch (InterruptedException e) {
			}
		}
		progressDialog.dispose();

		try {
			transFile.close();
		} catch (IOException e) {
		}

		if (receivedAll && !cancelled) {
			if (destFile.exists())
				destFile.delete();
			tmpFile.renameTo(destFile);

			new DushiOkDialog(parent.parentWindow, parent.strings.get(
					thisClass, "filetrans"), (parent.strings.get(thisClass,
					"filetrans") + " " + fileName + " " + parent.strings.get(
					thisClass, "completed")), parent.strings.get("ok"), false)
					.setVisible(true);
		} else {
			tmpFile.delete();

			new DushiOkDialog(parent.parentWindow, parent.strings.get(
					thisClass, "filetrans"), (parent.strings.get(thisClass,
					"filetrans") + " " + fileName + " " + parent.strings.get(
					thisClass, "cancelled")), parent.strings.get("ok"), false)
					.setVisible(true);
		}
	}

	protected void notifyAccepted(DushiUser user, boolean acc) {
		// 当我们是发送者时，此函数告诉用户是否接受了传输文件的邀请。
		senderDialog.accepted(user, acc);
	}

	protected void cancel() {
		// 终止文件传输。
		cancelled = true;
	}

	class SenderDialog extends Dialog implements ActionListener, KeyListener,
			WindowListener {
		private static final long serialVersionUID = 220L;
		private DushiPanel parent = null;
		private DushiUser[] users = null;
		private Label progLabel = null;
		private Label[] userLabels = null;
		private Button dismissButton = null;

		public SenderDialog(DushiPanel p, String fn, DushiUser[] u) {
			super(p.parentWindow, false);

			parent = p;
			fileName = fn;
			users = u;

			setTitle(parent.strings.get(DushiClientFileTransfer.class,
					"filetrans"));

			GridBagLayout myLayout = new GridBagLayout();
			setLayout(myLayout);

			Panel panel = new Panel();
			panel.setLayout(myLayout);

			panel.add(new Label(fileName), new GridBagConstraints(0, 0, 1, 1,
					1.0, 1.0, GridBagConstraints.NORTHWEST,
					GridBagConstraints.NONE, new Insets(0, 0, 0, 0), 0, 0));

			progLabel = new Label(parent.strings.get(
					DushiClientFileTransfer.class, "sending"));
			panel.add(progLabel, new GridBagConstraints(0, 1, 1, 1, 1.0, 1.0,
					GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
					new Insets(0, 0, 0, 0), 0, 0));

			userLabels = new Label[users.length];
			for (int count = 0; count < users.length; count++) {
				userLabels[count] = new Label(users[count].name
						+ ": "
						+ parent.strings.get(DushiClientFileTransfer.class,
								"waiting"));
				panel.add(userLabels[count], new GridBagConstraints(0,
						(count + 2), 1, 1, 1.0, 1.0,
						GridBagConstraints.NORTHWEST, GridBagConstraints.NONE,
						new Insets(0, 0, 0, 0), 0, 0));
			}

			dismissButton = new Button(parent.strings.get("dismiss"));
			dismissButton.addActionListener(this);
			dismissButton.addKeyListener(this);
			panel.add(dismissButton, new GridBagConstraints(0,
					(users.length + 2), 1, 1, 1.0, 1.0,
					GridBagConstraints.CENTER, GridBagConstraints.NONE,
					new Insets(0, 0, 0, 0), 0, 0));

			add(panel, new GridBagConstraints(0, 0, 1, 1, 1.0, 1.0,
					GridBagConstraints.NORTHWEST, GridBagConstraints.BOTH,
					new Insets(5, 5, 5, 5), 0, 0));

			pack();

			if (parent.parentWindow instanceof DushiWindow)
				DushiPanel.centerDialog(parent.parentWindow, this);
			else
				DushiPanel.centerDialogOnScreen(this);

			addWindowListener(this);
		}

		protected void progress(int prog) {
			if (!isVisible())
				return;

			String progText = parent.strings.get(DushiClientFileTransfer.class,
					"sending") + ": ";

			if (prog < 100)
				progText += prog + "%";
			else
				progText += parent.strings.get(DushiClientFileTransfer.class,
						"completed");
			progLabel.setText(progText);
			pack();
		}

		protected void accepted(DushiUser user, boolean acc) {
			if (!isVisible())
				return;

			String accString = user.name + ": ";

			if (acc)
				accString += parent.strings.get(DushiClientFileTransfer.class,
						"accepted");
			else
				accString += parent.strings.get(DushiClientFileTransfer.class,
						"refused");

			for (int count = 0; count < users.length; count++)
				if (users[count] == user) {
					userLabels[count].setText(accString);
					break;
				}
			pack();
		}

		public void actionPerformed(ActionEvent E) {
			if (E.getSource() == dismissButton)
				dispose();
		}

		public void keyPressed(KeyEvent E) {
		}

		public void keyReleased(KeyEvent E) {
			if ((E.getKeyCode() == KeyEvent.VK_ENTER)
					&& (E.getSource() == dismissButton))
				dispose();
		}

		public void keyTyped(KeyEvent E) {
		}

		public void windowActivated(WindowEvent E) {
		}

		public void windowClosed(WindowEvent E) {
		}

		public void windowClosing(WindowEvent E) {
			dispose();
		}

		public void windowDeactivated(WindowEvent E) {
		}

		public void windowDeiconified(WindowEvent E) {
		}

		public void windowIconified(WindowEvent E) {
		}

		public void windowOpened(WindowEvent E) {
		}
	}
}
